/*
 * led.c
 */

#include "em_device.h"
#include "em_chip.h"
#include "em_timer.h"

#include "tools.h"

#include "led.h"
#include "process.h"

#include "../hw/shiftreg.h"
#include "../hw/lightsense.h"

#include "../rtc/time.h"


/* led shiftreg state */
uint8_t led_sr_state[4];


/* led original values, 10-bit level (0-1023) per channel */
uint16_t led_eyes_set[2][3];
uint16_t led_skull_set[8];
uint16_t led_bone_set[5][3];

/* led actual target values */
uint16_t led_eyes[2][3];
uint16_t led_skull[8];
uint16_t led_bone[5][3];
uint16_t led_clock[16];

/* led scaling values */
uint16_t led_pwm_max;

const uint8_t led_eyes_scalemax[5] = {20, 100, 224, 255, 0};
const uint8_t led_skull_scalemax[5] = {20, 100, 224, 255, 0};
const uint8_t led_bone_scalemax[5] = {20, 100, 224, 255, 0};
const uint8_t led_clock_scalemax[5] = {16, 80, 200, 255, 255};

/* display disable */
uint8_t led_disable;


/* setup */
void led_init()
{
	uint16_t top;

	// set initial LED lines
	led_sr_state[LED_EYES_IDX] = 2;
	led_sr_state[LED_SKULL_IDX] = 3;
	led_sr_state[LED_BONE_IDX] = 4;
	led_sr_state[LED_CLOCK_IDX] = 7;

	// set up PWMs
	top = 0xffff;
	TIMER_TopSet(TIMER0, top);
	TIMER_TopSet(TIMER1, top);
	TIMER_TopSet(TIMER2, top);

	// set top reference value for scaler
	led_pwm_max = (SystemCoreClockGet() / 2000) - 1;
}

void led_stop()
{
	// set all PWMs inactive
	TIMER0->ROUTE &= ~(TIMER_ROUTE_CC0PEN | TIMER_ROUTE_CC1PEN | TIMER_ROUTE_CC2PEN);
	TIMER1->ROUTE &= ~(TIMER_ROUTE_CC0PEN | TIMER_ROUTE_CC1PEN | TIMER_ROUTE_CC2PEN);
	TIMER2->ROUTE &= ~(TIMER_ROUTE_CC0PEN | TIMER_ROUTE_CC1PEN | TIMER_ROUTE_CC2PEN);

	TIMER_Enable(TIMER0, false);

	// set all LED shift register outputs high/low as appropriate
	sr_dat[0] |=  LED_EYES_MASK;
	sr_dat[0] |=  LED_SKULL_MASK;
	sr_dat[0] &= ~LED_BONE_MASK;
	sr_dat[1] = 0;

	sr_update();
}


/* setting */


/* updating */
void led_state_next()
{
	uint16_t w;
	uint16_t x;
	uint16_t cmp[9];


	/* MATRIX SELECTS */
	// set eyes (RGB select)
	w = led_sr_state[LED_EYES_IDX]; 			// hardware is in BRG mode, we save the value
	led_sr_state[LED_EYES_IDX] += 2; 			// two ahead to get us into an RGB lookup.
	led_sr_state[LED_EYES_IDX] %= 3;
	sr_dat[0] &= ~LED_EYES_MASK;

	x = LED_EYES_MASK & ~((1 << w) << LED_EYES_SHIFT);
	sr_dat[0] |= x;

	// set skull (matrix select)
	led_sr_state[LED_SKULL_IDX]++;
	led_sr_state[LED_SKULL_IDX] %= 4;
	w = led_sr_state[LED_SKULL_IDX];
	sr_dat[0] &= ~LED_SKULL_MASK;				// \/ first bit separated from other 3

	x = (w == 0) ? 1 : ((1 << w) << LED_SKULL_SHIFT_HI);
	x |= (w == 1) ? 0x200 : 0; 					// << second bit has two locations
	x ^= LED_SKULL_MASK; 						// invert bits, low = on

	sr_dat[0] |= x;

	// set bone (led select)
	++led_sr_state[LED_BONE_IDX];
	led_sr_state[LED_BONE_IDX] %= 5;
	w = led_sr_state[LED_BONE_IDX];

	sr_dat[0] &= ~LED_BONE_MASK;
	sr_dat[0] |= ((1 << w) << LED_BONE_SHIFT); 	// this one's pretty normal thankfully

	// rotate clock (matrix select)
	led_sr_state[LED_CLOCK_IDX]++;
	led_sr_state[LED_CLOCK_IDX] %= 8;
	w = (1 << led_sr_state[LED_CLOCK_IDX]);

	// todo: find out wtf this is and remove it because it isn't anything
	/*if (w & 0x22) w ^= 0x22;
	if (w & 0x14) w ^= 0x14;
	*/
	sr_dat[1] = ~w;


	/* PWM LEVEL */
	w = led_sr_state[LED_EYES_IDX];
	cmp[LED_PWM_EYE_LEFT]  = led_eyes[0][w];
	cmp[LED_PWM_EYE_RIGHT] = led_eyes[1][w];

	w = led_sr_state[LED_SKULL_IDX];
	cmp[LED_PWM_SKULL_0] = led_skull[w];
	cmp[LED_PWM_SKULL_1] = led_skull[w + 4];

	w = led_sr_state[LED_BONE_IDX];
	cmp[LED_PWM_BONE_RED]   = led_bone[w][0];
	cmp[LED_PWM_BONE_GREEN] = led_bone[w][1];
	cmp[LED_PWM_BONE_BLUE]  = led_bone[w][2];

	w = led_sr_state[LED_CLOCK_IDX];
	cmp[LED_PWM_CLOCK_0] = led_clock[w];
	cmp[LED_PWM_CLOCK_1] = led_clock[w + 8];


	/* UPDATE AND ENABLE */
	// disconnect PWM pins
	TIMER0->ROUTE &= ~(TIMER_ROUTE_CC0PEN | TIMER_ROUTE_CC1PEN | TIMER_ROUTE_CC2PEN);
	TIMER1->ROUTE &= ~(TIMER_ROUTE_CC0PEN | TIMER_ROUTE_CC1PEN | TIMER_ROUTE_CC2PEN);
	TIMER2->ROUTE &= ~(TIMER_ROUTE_CC0PEN | TIMER_ROUTE_CC1PEN | TIMER_ROUTE_CC2PEN);

	// set new compare values
	TIMER_CompareSet(TIMER0, 0, cmp[0]);
	TIMER_CompareSet(TIMER0, 1, cmp[1]);
	TIMER_CompareSet(TIMER0, 2, cmp[2]);
	TIMER_CompareSet(TIMER1, 0, cmp[3]);
	TIMER_CompareSet(TIMER1, 1, cmp[4]);
	TIMER_CompareSet(TIMER1, 2, cmp[5]);
	TIMER_CompareSet(TIMER2, 0, cmp[6]);
	TIMER_CompareSet(TIMER2, 1, cmp[7]);
	TIMER_CompareSet(TIMER2, 2, cmp[8]);

	// clock out shift register data
	sr_update();

	// re-enable PWMs
	// spent hours on this one. basically, even though we updated the non-buffered
	// compare values above, they will NOT be used until the timer rolls over.
	// according to datasheet, this is what BUFFERED value should do, but we don't use that...
	// so, we set cnt to max so that next clock cycle, it will have rolled over and PWMs update.
	TIMER0->CNT = TIMER1->CNT = TIMER2->CNT = 0xffe0;
	TIMER0->ROUTE |= (TIMER_ROUTE_CC0PEN | TIMER_ROUTE_CC1PEN | TIMER_ROUTE_CC2PEN);
	TIMER1->ROUTE |= (TIMER_ROUTE_CC0PEN | TIMER_ROUTE_CC1PEN | TIMER_ROUTE_CC2PEN);
	TIMER2->ROUTE |= (TIMER_ROUTE_CC0PEN | TIMER_ROUTE_CC1PEN | TIMER_ROUTE_CC2PEN);
}

/* rendering */
static uint16_t led_get_scale(const uint8_t *scalemax)
{
	int i;
	uint32_t scale;

	// scale brightness
	// first find our scale range
	for (i = 3; i; i--) {
		if (lsens_gain < lsens_ref[i]) break;
	}

	// then get our scale percentage
	scale = scale16i(lsens_gain,
				lsens_ref[i + 1], lsens_ref[i],
				scalemax[i + 1], scalemax[i]);
	scale *= led_pwm_max;
	scale >>= 8;

	return (uint16_t)scale;
}

void led_render_eyes()
{
	led_eyes[0][0] = scale16ufaster(led_eyes_set[0][0], 10, led_get_scale(led_eyes_scalemax));
	led_eyes[0][1] = scale16ufaster(led_eyes_set[0][1], 10, led_get_scale(led_eyes_scalemax));
	led_eyes[0][2] = scale16ufaster(led_eyes_set[0][2], 10, led_get_scale(led_eyes_scalemax));

	led_eyes[1][0] = scale16ufaster(led_eyes_set[0][0], 10, led_get_scale(led_eyes_scalemax));
	led_eyes[1][1] = scale16ufaster(led_eyes_set[0][1], 10, led_get_scale(led_eyes_scalemax));
	led_eyes[1][2] = scale16ufaster(led_eyes_set[0][2], 10, led_get_scale(led_eyes_scalemax));
}

void led_render_skull()
{
	int i;

	for (i = 0; i < 8; i++) {
		led_skull[i] = scale16ufaster(led_skull_set[i], 10, led_get_scale(led_skull_scalemax));
	}
}

void led_render_bone()
{
	int i;

	for (i = 0; i < 5; i++) {
		led_bone[i][0] = scale16ufaster(led_bone_set[i][0], 10, led_get_scale(led_bone_scalemax));
		led_bone[i][1] = scale16ufaster(led_bone_set[i][1], 10, led_get_scale(led_bone_scalemax));
		led_bone[i][2] = scale16ufaster(led_bone_set[i][2], 10, led_get_scale(led_bone_scalemax));
	}
}

void led_render_clock()
{
	int i;

	uint16_t w;
	uint16_t low;

	if ((led_disable & (1 << LED_CLOCK_IDX)) == 0) {
		w = rtc_binclock_render(RTC_CLOCK_RENDER_LEFT);
		low = 20;
	} else {
		w = 0;
		low = 0;
	}

	for (i = 0; i < 16; i++) {
		led_clock[i] = (w >> i) & 1 ? 1024 : low;
	}

	// scale the LEDs
	for (i = 0; i < 16; i++) {
		led_clock[i] = scale16ufaster(led_clock[i], 10, led_get_scale(led_clock_scalemax));
	}
}
