/**
 * osc.c
 */

#include "em_device.h"
#include "em_chip.h"
#include "em_cmu.h"

#include "tools.h"


void hfclk_set(int band)
{
	// disable systick interrupt
	NVIC_DisableIRQ(SysTick_IRQn);

	// change clock frequency
	CMU_HFRCOBandSet(band); 	// cmuHFRCOBand_14MHz
	CMU_ClockDivSet(cmuClock_HF, 1);
}

void hfclk_cal(uint32_t cycles)
{
	/* run the calibration routine to know the initial value of the up counter */
	uint32_t upctr = CMU_Calibrate(cycles, cmuOsc_LFXO);

	/* expected tuning value, based on math: cycles x LFCLK_freq[32768] / HFCLK_freq */
	uint32_t upctr_tuned = (cycles * 32768) / SystemCoreClockGet();

	/* Variable to store the previous and tuned upcounter value for comparison */
	uint32_t upctr_prev;

	/* Read the initial tuning value */
	uint8_t  tune = CMU_OscillatorTuningGet(cmuOsc_HFRCO);

	/* If the up counter result is smaller than the tuned value, the HFRCO is running
	 * at a higher frequency so the HFRCO tuning register has to be decremented */
	if (upctr < upctr_tuned) {
		while (upctr < upctr_tuned) {
			/* store the previous value for comparison */
			upctr_prev = upctr;
			/* Decrease tuning value */
			tune--;
			/* Write the new HFRCO tuning value */
			CMU_OscillatorTuningSet(cmuOsc_HFRCO, tune);
			/* Run the calibration routine again */
			upctr = CMU_Calibrate(cycles, cmuOsc_LFXO);
		}

		/* Check which value goes closer to the tuned one and increase the tuningval
		 * if the previous value is closer */
		if ((upctr - upctr_tuned) > (upctr_tuned - upctr_prev)) {
			tune++;
		}
	}

	/* If the up counter result is higher than the desired value, the HFRCO is running
	 * at a lower frequency so the tuning value has to be incremented */
	if (upctr > upctr_tuned) {
		while (upctr > upctr_tuned) {
			/* store the previous value for comparison */
			upctr_prev = upctr;
			/* Increase tuning value */
			tune++;
			/* Write the new HFRCO tuning value */
			CMU_OscillatorTuningSet(cmuOsc_HFRCO, tune);
			/* Run the calibration routine again */
			upctr = CMU_Calibrate(cycles, cmuOsc_LFXO);
		}

		/* Check which value goes closer to the tuned one and decrease the tuningval
		 * if the previous value is closer */
		if ((upctr_tuned - upctr) > (upctr_prev - upctr_tuned)) {
			tune--;
		}
	}
}

void __attribute__((optimize("-O0"))) osc_init(uint32_t band)
{
	// set CPU to low frequency while lfxo starts up
	hfclk_set(cmuHFRCOBand_1MHz);

	// start lfxo
	CMU_OscillatorEnable(cmuOsc_LFXO, true, true);

	// set CPU clock to standard operating speed, then calibrate
	// by now our LFXO is up and running so we can calibrate against it
	hfclk_set(band);
	//hfclk_cal(0x1ffff); //0x3ffff);
	CMU->HFRCOCTRL |= 0xff; 	// set highest speed in this band
}
