/*
 * spiflash.c
 * efm32 driver for winbond w25q32fv series flash
 * ---
 * todo:
 * [ ] support more flashtypes
 * [ ] don't do actions if no flash was found
 * [ ] make this universal and not tied to EFM32
 */

#include <stdint.h>

#include "em_device.h"
#include "em_chip.h"
#include "em_cmu.h"
#include "em_dma.h"
#include "em_usart.h"

#include "../lib/dma/dmactrl.h"

#include "spiflash.h"


/* prototypes */
static void spiflash_cs(uint8_t active);


/* models */
static const SPIFLASH_MODEL spiflash_model[] = {
	{
		SPIFLASH_VEND_WINBOND,
		SPIFLASH_MODEL_W25Q32,
		"Winbond W25Q32FV"
	}
};


/* callback */
static DMA_CB_TypeDef 	spiflash_cb;


/* dma */
#define DMA_CHANNEL_RX   0
#define DMA_CHANNEL_TX   1
#define DMA_CHANNELS     2

static volatile uint8_t dma_rx_busy;
static volatile uint8_t dma_tx_busy;

static 			void    *tx_buf;
static 			uint16_t tx_len;
static volatile uint32_t tx_len_remain;
static 			uint32_t tx_next_addr;

static 			void   (*rx_done_cb)(void);
static 			void   (*tx_done_cb)(void);


// system-wide scratch buffer
uint8_t spiflash_buf[256];


/* dma functions */
static void spi_dma_done(unsigned int chan, bool primary, void *userfn)
{
	(void) primary;
	void (*fn)();

	// Clear flag to indicate complete transfer
	if (chan == DMA_CHANNEL_RX) {
		dma_rx_busy = false;
	} else if (chan == DMA_CHANNEL_TX) {
		dma_tx_busy = false;
	}

	// completely done?
	if (!(dma_tx_busy | dma_rx_busy)) {
		// deassert flash
		spiflash_cs(0);

		// user-provided completion callback
		if (userfn != 0) {
			fn = (void (*)())userfn;
			fn();
		}
	}
}

static inline void spi_dma_init(void)
{
	// dma init structs */
	DMA_Init_TypeDef        dmainit;
	DMA_CfgChannel_TypeDef  rx_ch_cfg;
	DMA_CfgDescr_TypeDef    rx_desc_cfg;
	DMA_CfgChannel_TypeDef  tx_ch_cfg;
	DMA_CfgDescr_TypeDef    tx_desc_cfg;

	// Initializing the DMA
	dmainit.hprot        = 0;
	dmainit.controlBlock = dmaControlBlock;
	DMA_Init(&dmainit);

	// Setup call-back function
	spiflash_cb.cbFunc  = spi_dma_done;
	spiflash_cb.userPtr = NULL;


	/*** Setting up RX DMA ***/
	// Setting up channel
	rx_ch_cfg.highPri   = false;
	rx_ch_cfg.enableInt = true;
	rx_ch_cfg.select    = SPIFLASH_DMA_RXDATAV;
	rx_ch_cfg.cb        = &spiflash_cb;
	DMA_CfgChannel(DMA_CHANNEL_RX, &rx_ch_cfg);

	// Setting up channel descriptor
	rx_desc_cfg.dstInc  = dmaDataInc1;
	rx_desc_cfg.srcInc  = dmaDataIncNone;
	rx_desc_cfg.size    = dmaDataSize1;
	rx_desc_cfg.arbRate = dmaArbitrate1;
	rx_desc_cfg.hprot   = 0;
	DMA_CfgDescr(DMA_CHANNEL_RX, true, &rx_desc_cfg);


	/*** Setting up TX DMA ***/
	// Setting up channel
	tx_ch_cfg.highPri   = false;
	tx_ch_cfg.enableInt = true;
	tx_ch_cfg.select    = SPIFLASH_DMA_TXBL;
	tx_ch_cfg.cb        = &spiflash_cb;
	DMA_CfgChannel(DMA_CHANNEL_TX, &tx_ch_cfg);

	// Setting up channel descriptor
	tx_desc_cfg.dstInc  = dmaDataIncNone;
	tx_desc_cfg.srcInc  = dmaDataInc1;
	tx_desc_cfg.size    = dmaDataSize1;
	tx_desc_cfg.arbRate = dmaArbitrate1;
	tx_desc_cfg.hprot   = 0;
	DMA_CfgDescr(DMA_CHANNEL_TX, true, &tx_desc_cfg);
}


/* flash handler */
void spiflash_init()
{
	// start USART clock
	SPIFLASH_CMUCLK_EN(true);

	// configure USART
	// this is done at system init for us already, so we don't need to do this now

	// set up DMA
	spi_dma_init();

	// stop USART clock
	SPIFLASH_CMUCLK_EN(false);
}

static inline void spiflash_cs(uint8_t active)
{
	if (active) {
		SPIFLASH_CMUCLK_EN(true);
		GPIO->P[SPIFLASH_CS_PORT].DOUTCLR = SPIFLASH_CS_PIN;
	} else {
		GPIO->P[SPIFLASH_CS_PORT].DOUTSET = SPIFLASH_CS_PIN;
		SPIFLASH_CMUCLK_EN(false);
	}
}

/*
 * blocking spi read/write routine
 * assumes you set chipsel properly
 */
static uint8_t spi_rw(uint8_t byte)
{
	// send byte
	SPIFLASH_SPI->TXDATA = byte;
	// wait while sending
	while (!(SPIFLASH_SPI->STATUS & USART_STATUS_TXC));
	// return received data
	return SPIFLASH_SPI->RXDATA;
}

/*
 * non-blocking spi read/write routine using DMA
 * use null for tx or rx if you don't need it
 * assumes you set chipsel properly
 */
static SPIFLASH_STATUS spi_rw_dma(uint8_t *tx, uint8_t *rx, int bytes)
{
	if (dma_tx_busy || dma_rx_busy) return SPIFLASH_FAIL_DMA_BUSY;
	if (!((uint32_t)tx | (uint32_t)rx)) return SPIFLASH_FAIL;

	// Only activate RX DMA if a receive buffer is specified
	if (rx) {
		// Setting flag to indicate that RX is in progress
		// will be cleared by call-back function
		dma_rx_busy = 1;

		// Clear USART RX registers
		SPIFLASH_SPI->CMD = USART_CMD_CLEARRX;

		// Activate RX channel
		DMA_ActivateBasic(DMA_CHANNEL_RX, true, false,
				rx, (void *)&(SPIFLASH_SPI->RXDATA), bytes - 1);

		// tx buffer needs to be pointing _somehwere_ for this to work
		if (!tx) {
			tx = rx;
		}
	}

	// Setting flag to indicate that TX is in progress
	// will be cleared by call-back function
	dma_tx_busy = true;

	// Clear TX regsiters
	SPIFLASH_SPI->CMD = USART_CMD_CLEARTX;

	// Activate TX channel
	DMA_ActivateBasic(DMA_CHANNEL_TX, true, false,
			(void *)&(SPIFLASH_SPI->TXDATA), tx, bytes - 1);

	return SPIFLASH_DMA_STARTED;
}

uint8_t spiflash_is_busy()
{
	return (spiflash_get_reg(SPIFLASH_CMD_RD_SR1) & SPIFLASH_SR1_BUSY);
}

static inline SPIFLASH_STATUS spiflash_is_busy_wait()
{
	uint16_t timeout = SPIFLASH_DEF_TIMEOUT;

	while (spiflash_is_busy() && timeout--);
	if (!timeout) return SPIFLASH_TIMEOUT;

	return SPIFLASH_OK;
}

uint8_t spiflash_get_reg(uint8_t reg)
{
	uint8_t ret;

	spiflash_cs(true);

	spi_rw(reg);
	ret = spi_rw(0xff);

	spiflash_cs(false);

	return ret;
}

void spiflash_set_reg(uint8_t reg)
{
	spiflash_cs(true);

	spi_rw(reg);

	spiflash_cs(false);
}

SPIFLASH_STATUS spiflash_get_id(uint8_t *mfr, uint16_t *dev)
{
	uint16_t id2;

	spiflash_cs(true);

	spi_rw(SPIFLASH_CMD_JEDEC_ID);
	// spi_rw(0x00); spi_rw(0x00); spi_rw(0x00);

	*mfr = spi_rw(0xff);
	id2  = spi_rw(0xff) << 8;
	id2 |= spi_rw(0xff);

	*dev = id2;

	spiflash_cs(false);

	return SPIFLASH_OK;
}

SPIFLASH_MODEL * spiflash_get_model()
{
	uint8_t mfr;
	uint16_t dev;
	int i;

	spiflash_get_id(&mfr, &dev);

	for (i = 0; i < 1; i++) {
		if (spiflash_model[i].manufacturer == mfr && spiflash_model[i].jedec == dev) {
			return (SPIFLASH_MODEL *)&spiflash_model[i];
		}
	}

	return (SPIFLASH_MODEL *)0;
}

void spiflash_start_prog_page(uint32_t addr)
{
	// disable write protect
	spiflash_set_reg(SPIFLASH_CMD_WREN);

	// enable flash and send address
	spiflash_cs(true);

	spi_rw(SPIFLASH_CMD_PROG_PAGE);
	spi_rw((addr >> 16) & 0xff);
	spi_rw((addr >>  8) & 0xff);
	spi_rw((addr      ) & 0xff);
}

void spiflash_readbuf_cb()
{
	if (rx_done_cb != 0) {
		rx_done_cb();
		rx_done_cb = 0;
	}
}

SPIFLASH_STATUS spiflash_readbuf(void *buf, uint32_t addr, uint32_t bytes, void (*done_cb)())
{
	// wait until flash is ready
	if (done_cb && (dma_tx_busy || dma_rx_busy)) return SPIFLASH_FAIL_DMA_BUSY;
	if (spiflash_is_busy_wait()) return SPIFLASH_TIMEOUT;

	// enable flash and send command
	spiflash_cs(true);

	spi_rw(SPIFLASH_CMD_READ);
	spi_rw((addr >> 16) & 0xff);
	spi_rw((addr >>  8) & 0xff);
	spi_rw((addr      ) & 0xff);

	// work the data
	// non-blocking with DMA
	if (done_cb) {
		// set up our callback handler
		spiflash_cb.userPtr = spiflash_readbuf_cb;

		// set up our callback and buffer handlers
		rx_done_cb = done_cb;

		return spi_rw_dma(0, buf, bytes);
	}

	// blocking
	uint8_t *b = (uint8_t *)buf;

	while (bytes--) {
		*b++ = spi_rw(0xff);
	}

	spiflash_cs(false);

	return SPIFLASH_OK;
}

static void spiflash_writebuf_cb()
{
	spiflash_cs(false);

	if (tx_len_remain) {
		// set up transfer size
		if (tx_len_remain > 256) {
			tx_len = 256;
			tx_len_remain -= 256;
		} else {
			tx_len = tx_len_remain;
			tx_len_remain = 0;
		}

		// start up spi, send command and address
		spiflash_start_prog_page(tx_next_addr);

		// set next address
		tx_next_addr += tx_len;

		// set up new transfer
		spi_rw_dma(tx_buf, 0, tx_len);
	} else {
		if (tx_done_cb != 0) {
			tx_done_cb();
			tx_done_cb = 0;
		}
	}
}

SPIFLASH_STATUS spiflash_writebuf(void *buf, uint32_t addr, uint32_t bytes, void (*done_cb)())
{
	uint16_t tx_size;

	// wait until flash is ready
	if (done_cb && (dma_tx_busy || dma_rx_busy)) return SPIFLASH_FAIL_DMA_BUSY;
	if (spiflash_is_busy_wait()) return SPIFLASH_TIMEOUT;

	// figure out max transmit size based on block size
	tx_size = 256 - (addr & 0xff);

	// disable write protect
	spiflash_set_reg(SPIFLASH_CMD_WREN);

	// start up spi, send command and address
	spiflash_start_prog_page(addr);

	// non-blocking with DMA
	if (done_cb) {
		// set up our callback handler
		spiflash_cb.userPtr = spiflash_writebuf_cb;

		// set up our callback and buffer handlers
		tx_done_cb = done_cb;
		tx_buf = buf;
		if (bytes > tx_size) {
			tx_len_remain = bytes - tx_size;
			bytes = tx_size;
		} else {
			tx_len_remain = 0;
		}

		tx_next_addr = addr + bytes;

		return spi_rw_dma(tx_buf, 0, bytes);
	}

	// blocking
	uint8_t *b = (uint8_t *)buf;
	uint16_t remain;

	if (bytes > tx_size) {
		remain = bytes - tx_size;
		bytes = tx_size;
	} else {
		remain = 0;
	}

	while (bytes) {
		// set next possible address
		addr += bytes;

		while (bytes--) {
			spi_rw(*b++);
		}

		spiflash_cs(false);

		if (remain > 256) {
			bytes = 256;
			remain -= 256;
		} else {
			bytes = remain;
			remain = 0;
		}

		if (bytes) {
			// wait until flash is done
			if (spiflash_is_busy_wait()) return SPIFLASH_TIMEOUT;

			// start up spi, send command and address
			spiflash_start_prog_page(addr);
		}
	}

	return SPIFLASH_OK;
}

SPIFLASH_STATUS spiflash_erasepage(uint32_t addr, uint8_t pagesize_k)
{
	uint8_t cmd;

	switch (pagesize_k) {
		case 4: 	{ cmd = SPIFLASH_CMD_ERASE_SECTOR; break; }
		case 32: 	{ cmd = SPIFLASH_CMD_ERASE_BLOCK32; break; }
		case 64: 	{ cmd = SPIFLASH_CMD_ERASE_BLOCK64; break; }
		default: 	{ return SPIFLASH_INVALID_OPTION; }
	}

	// disable write protect
	spiflash_set_reg(SPIFLASH_CMD_WREN);

	// erase the page
	spiflash_cs(true);

	spi_rw(cmd);
	spi_rw((addr >> 16) & 0xff);
	spi_rw((addr >> 8)  & 0xff);
	spi_rw(addr         & 0xff);

	spiflash_cs(false);

	return SPIFLASH_OPERATION_STARTED;
}

SPIFLASH_STATUS spiflash_erasechip(uint32_t confirm)
{
	if (confirm == 0x0) {

	}

	return SPIFLASH_FAIL;
}
