/**
 * spiflash.h
 */

#ifndef __WP2016_HW_SPIFLASH_H
#define __WP2016_HW_SPIFLASH_H


#include "em_cmu.h"


// device types
typedef struct {
	uint8_t   manufacturer;
	uint16_t  jedec;
	char     *description;
} SPIFLASH_MODEL;

// return types
typedef enum {
	SPIFLASH_OK,
	SPIFLASH_FAIL,
	SPIFLASH_FAIL_BUSY,
	SPIFLASH_FAIL_DMA_BUSY,
	SPIFLASH_FAIL_SIZE_TOO_LARGE,
	SPIFLASH_INVALID_OPTION,
	SPIFLASH_TIMEOUT,
	SPIFLASH_OPERATION_STARTED,
	SPIFLASH_DMA_STARTED,
} SPIFLASH_STATUS;

// settings
#define SPIFLASH_DEF_TIMEOUT 	1000


// spi peripheral setup
#define SPIFLASH_SPI 			USART1
#define SPIFLASH_CMUCLK_EN(en) 	CMU_ClockEnable(cmuClock_USART1, en)

#define SPIFLASH_DMA_RXDATAV 	DMAREQ_USART1_RXDATAV
#define SPIFLASH_DMA_TXBL 		DMAREQ_USART1_TXBL

#define SPIFLASH_DATARATE 		6000000 	// we will never be below 14mhz, max is HFPER/2

#define SPIFLASH_CS_PORT 		2 			// this module manages !cs manually. please deroute this
#define SPIFLASH_CS_PIN			(1 << 0) 	// pin from the USART and set as push/pull, default high


// identification
#define SPIFLASH_VEND_WINBOND 			0xef
#define SPIFLASH_MODEL_W25Q32 			0x4016
#define SPIFLASH_ERASESTATE_W25Q32 		0xff


// commands
#define SPIFLASH_CMD_WREN 				0x06 	// set prior to any program, erase, wrstatus
#define SPIFLASH_CMD_WREN_VSR 			0x50 	// set instead of WREN prior to volatile SR change
#define SPIFLASH_CMD_WRDIS 				0x04
#define SPIFLASH_CMD_RD_SR1 			0x05 	// s7-s0
#define SPIFLASH_CMD_WR_SR1 			0x01 	// can program status reg 1 and 2
#define SPIFLASH_CMD_RD_SR2 			0x35 	// s15-s8
#define SPIFLASH_CMD_WR_SR2 			0x31
#define SPIFLASH_CMD_RD_SR3 			0x15 	// s23-s16
#define SPIFLASH_CMD_WR_SR3 			0x11
#define SPIFLASH_CMD_ERASE_CHIP 		0xc7
#define SPIFLASH_CMD_EP_SUSPEND 		0x75
#define SPIFLASH_CMD_EP_RESUME 			0x7a
#define SPIFLASH_CMD_PWRDOWN 			0xb9
#define SPIFLASH_CMD_PWRUP 				0xab 	// ?? ?? ?? id7-id0
#define SPIFLASH_CMD_MANU_DEVID 		0x90 	// ?? ?? 00 mf7-mf0 id7-id0
#define SPIFLASH_CMD_JEDEC_ID 			0x9f 	// mf7-mf0 id15-id8 id7-id0
#define SPIFLASH_CMD_GLOBAL_LOCK 		0x7e
#define SPIFLASH_CMD_GLOBAL_UNLOCK 		0x98
#define SPIFLASH_CMD_RESET_ENABLE 		0x66
#define SPIFLASH_CMD_RESET 				0x99
#define SPIFLASH_CMD_UNIQEID 			0x4b 	// ?? ?? ?? ?? UID
#define SPIFLASH_CMD_PROG_PAGE 			0x02 	// a23-a16 a15-a8 a7-a0 d7-d0 ... (up to 256)
#define SPIFLASH_CMD_ERASE_SECTOR 		0x20 	// address, 4K size
#define SPIFLASH_CMD_ERASE_BLOCK32 		0x52 	// address, 32k size
#define SPIFLASH_CMD_ERASE_BLOCK64 		0xd8 	// address, 64k size
#define SPIFLASH_CMD_READ 				0x03 	// address (data)
#define SPIFLASH_CMD_FASTREAD 			0x0b 	// address ?? (data)

// constants
#define SPIFLASH_SR1_BUSY 				(1 << 0) 	// erase or write in progress (ro)
#define SPIFLASH_SR1_WEL 				(1 << 1) 	// write enable latch (ro)
#define SPIFLASH_SR1_BP 				(7 << 2) 	// block protect
#define SPIFLASH_SR1_TB 				(1 << 5) 	// top/bottom protect
#define SPIFLASH_SR1_SEC 				(1 << 6) 	// sector protect
#define SPIFLASH_SR1_SRP0 				(1 << 7) 	// status register protect 0

#define SPIFLASH_SR2_SRP1 				(1 << 0) 	// status register protect 1
#define SPIFLASH_SR2_QE 				(1 << 1) 	// quad enable
#define SPIFLASH_SR2_LB 				(7 << 3) 	// sr lock bits
#define SPIFLASH_SR2_CMP 				(1 << 6) 	// complement protect
#define SPIFLASH_SR2_SUS 				(1 << 7) 	// suspend status (ro)

#define SPIFLASH_SR3_WPS 				(1 << 2) 	// write protect select
#define SPIFLASH_SR3_DRV 				(3 << 5)	// output driver strength
#define SPIFLASH_SR3_HOLD_RST 			(1 << 7) 	// hold/reset function select


// system-wide buffer
extern uint8_t spiflash_buf[256];


// prototypes
void spiflash_init();

uint8_t spiflash_is_busy();

uint8_t spiflash_get_reg(uint8_t reg);
void spiflash_set_reg(uint8_t reg);

SPIFLASH_STATUS spiflash_get_id(uint8_t *mfr, uint16_t *dev);

SPIFLASH_STATUS spiflash_readbuf(void *buf, uint32_t addr, uint32_t bytes, void (*done_cb)());
SPIFLASH_STATUS spiflash_writebuf(void *buf, uint32_t addr, uint32_t bytes, void (*done_cb)());

SPIFLASH_STATUS spiflash_erasepage(uint32_t addr, uint8_t pagesize_k);


#endif
