#include "em_device.h"
#include "em_chip.h"
#include "em_cmu.h"
#include "em_rtc.h"

#include "time.h"

#include "display/led.h"


uint8_t time_month 		= 1;
uint8_t time_day 		= 1;
uint8_t time_hour 		= 0;
uint8_t time_minute 	= 0;
uint8_t time_second 	= 0;

static uint8_t clock_is_set;
static char time_text[32];


/* months */
const char month[12][4] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
};


/* prototypes */
static uint8_t days_per_month(uint8_t month);


/* setup */
void rtc_interrupt_enable()
{
	RTC_IntClear(RTC_IFC_COMP0);
	RTC_IntEnable(RTC_IF_COMP0);

	NVIC_SetPriority(RTC_IRQn, 6);
	NVIC_EnableIRQ(RTC_IRQn);
}

void rtc_interrupt_disable()
{
	NVIC_DisableIRQ(RTC_IRQn);
	RTC_IntDisable(RTC_IFS_OF);
}

void rtc_init()
{
	// make sure RTC is stopped before proceeding
	RTC_Enable(false);

	// set prescaler to count every second
	CMU_ClockDivSet(cmuClock_RTC, cmuClkDiv_16384);

	// set rtc interrupt to fire every second
	RTC_CompareSet(0, 1);

	// turn on the RTC
	RTC_Enable(true);
}

uint8_t rtc_is_clock_set()
{
	return clock_is_set;
}


static inline void rtc_tick_minute()
{
	// minute elapsed
	time_minute++;
	if (time_minute >= 60) {
		// hour elapsed
		time_minute = 0;
		time_hour++;
		if (time_hour >= 24) {
			// day elapsed
			time_hour = 0;
			time_day++;
			if (time_day > days_per_month(time_month)) {
				// month elapsed
				time_month++;
				time_month %= 12;
			}
		}
	}
}

uint8_t rtc_tick()
{
	time_second++;
	if (time_second >= 60) {
		time_second = 0;
		rtc_tick_minute();
	}

	return time_second;
}


/* console and ui */
/*
 * returns the bitvalue needed for the clock.
 * flashes VCR style if the clock isn't set.
 */
static uint8_t clock_unset_render_state = 0;
uint16_t rtc_binclock_render(uint8_t align)
{
	// bit pattern:
	// 0 = low bit, high = high bit
	// <h3 h2> <m1 m2 m3 m4 m5 m0> <h1 h0> <s1 s2 s3 s4 s5 s0>

	uint16_t ret = 0;
	uint8_t hr;
	// uint8_t pm;

	if (!clock_is_set) {
		clock_unset_render_state++;
		clock_unset_render_state %= 16;

		switch (clock_unset_render_state) {
			case 0: { return 0x0141; break; }
			case 1: { return 0x21e1; break; }
			case 2: { return 0x71f1; break; }
			case 3: { return 0xf9f9; break; }
			case 4: { return 0xfdfd; break; }
			case 5:
			case 6:
			case 12:
			case 13:{ return 0xffff; break; }
			default: {
				if (clock_unset_render_state <= 11) {
					return ((clock_unset_render_state & 1) ? 0xffff : 0x0000);
				}
				return 0x0000;
			}
		}
	}

	switch (align) {
		case RTC_CLOCK_RENDER_LEFT: {
			// hours
			// pm = time_hour > 12 ? 1 : 0;
			hr = time_hour; if (time_hour > 12) hr -= 12;
			ret = (hr & 0x3) << 6; ret |= (hr & 0xc) << 12;

			// minutes
			ret |= (time_minute & 0x09) << 8;
			ret |= (time_minute & 0x02) << 12;
			ret |= (time_minute & 0x04) << 10;
			ret |= (time_minute & 0x10) << 6;
			ret |= (time_minute & 0x20) << 4;

			// minutes
			ret |= (time_second & 0x09);
			ret |= (time_second & 0x02) << 4;
			ret |= (time_second & 0x04) << 2;
			ret |= (time_second & 0x10) >> 2;
			ret |= (time_second & 0x20) >> 4;
			break;
		}

		case RTC_CLOCK_RENDER_RIGHT:
		default: {
			// if you want it to render to the right...
			// just turn the fucking badge upside down.
			break;
		}
	}

	return ret;
}

char * rtc_print_time()
{
	if (clock_is_set) {
		sprintf(time_text, "%d:%02d:%02d", time_hour, time_minute, time_second);
		return time_text;
	}

	return "Clock not set. Run 'help time' for instructions.";
}

char * rtc_print_date()
{
	if (clock_is_set) {
		sprintf(time_text, "%s %02d", month[time_month - 1], time_day);
		return time_text;
	}

	return "Clock not set. Run 'help date' for instructions.";
}

int rtc_set_time(uint8_t hour, uint8_t minute, uint8_t second)
{
	if (hour < 24 && minute < 60 && second < 60) {
		time_hour = hour;
		time_minute = minute;
		time_second = second;
		clock_is_set = 1;
		return 0;
	}

	return -1;
}

int rtc_set_date(uint8_t month, uint8_t day)
{
	if (month <= 12 && day <= 31) {
		// con won't run through the weekend
		// so every month has 31 days. rejoice
		time_month = month;
		time_day = day;
		return 0;
	}

	return -1;
}


/* helpers */
static uint8_t days_per_month(uint8_t month)
{
	switch (month) {
		case 1:
		case 3:
		case 5:
		case 7:
		case 9:
		case 10:
		case 12: {
			return 31;
		}
		case 4:
		case 6:
		case 8:
		case 11: {
			return 30;
		}
		case 2: {
			return 28;
		}
		default: {
			return 0;
		}
	}
}
